// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (MPU401.CC) Copyright 1998 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com




#include "bochs.h"


#if BX_USE_MPU401_SMF
bx_mpu401_c bx_mpu401;
#define this NULL
#endif

#define WRITELOG 1

bx_mpu401_c::bx_mpu401_c(void)
{
#if BX_MPU401_LOG > 0
  LOGFILE = fopen(BX_MPU401_LOGFILE,"a");	// logfile for errors etc.
#endif
  DATAFILE = fopen(BX_MPU401_DATA,"w");	// the midi stream is written to this file
  INBUF init( (int) INBUFLEN);		// the input
  OUTBUF init( (int) OUTBUFLEN);	// and output buffers
  BX_MPU401_THIS s.uartmode = 0;	// only generate a stream in uart mode
}

bx_mpu401_c::~bx_mpu401_c(void)
{
#if BX_MPU401_LOG > 0
  fclose(LOGFILE);
#endif
  fclose(DATAFILE);
}

  void
bx_mpu401_c::init(bx_devices_c *d)
{
  BX_MPU401_THIS devices = d;

//#if 0
  /* MPU401 */

  writelog("driver initialised, IRQ %d, IO %03x", 
	BX_MPU401_IRQ, BX_MPU401_IO);

// the IRQ is not implemented yet, and probably not even necessary
//  BX_MPU401_THIS devices->register_irq(BX_MPU401_IRQ, "MPU 401");

// The MPU401 takes 4 IO addresses, although only the first two
// are actually in use
  for (unsigned addr=BX_MPU401_IO; addr<=BX_MPU401_IO+3; addr++) {
    BX_MPU401_THIS devices->register_io_read_handler(this,
       read_handler,
       addr, "MPU 401");
    BX_MPU401_THIS devices->register_io_write_handler(this,
       write_handler,
       addr, "MPU 401");
    }
//#endif
}



  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_mpu401_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_MPU401_SMF
  bx_mpu401_c *class_ptr = (bx_mpu401_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}


  Bit32u
bx_mpu401_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_MPU401_SMF
  UNUSED(address);
  UNUSED(io_len);
  Bit32u result = 0;

// we support only byte access to the port
  if (io_len != 1)
    {
    writelog("not byte access!");
    return(result);
    }

#if BX_MPU401_LOG > 1
  writelog("access read at %03x for %d", address, io_len);
#endif

  if (address == STATPORT)
    {	
// the status port shows if input or output are ready
// Note that the bits are inverse to their meaning
    if (INBUF full() != 0) result |= OUTPUT_READY;	// no more output
    if (OUTBUF empty() != 0) result |= INPUT_AVAIL;	// no input available
#if BX_MPU401_LOG > 1
    writelog("status port, result %02x", result);
#endif
    }
  else if (address == DATAPORT)
    {
// data port: contains an MPU_ACK after receiving a command
// Will contain other data as well when midi input is supported
    Bit8u res8bit;
    if ( OUTBUF get(&res8bit) == 0)
      writelog("data port not ready - no data in buffer");
    else
      result = (Bit32u) res8bit;
#if BX_MPU401_LOG > 1
    writelog("data port, result %02x", result);
#endif
    }
  else
    writelog("read from unsupported port!");

  /* MPU401 */
  return(result);
}

  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_mpu401_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_MPU401_SMF
  bx_mpu401_c *class_ptr = (bx_mpu401_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_mpu401_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_MPU401_SMF

  int success = 0;

#if BX_MPU401_LOG > 1
  writelog("access write at %03x for %d to %02x", address, io_len, value);
#endif

// we only support byte access to the prot
  if (io_len != 1)
	{
	writelog("not byte access!");
	return;
	}

  if (address == COMDPORT)
    {
// the command port: Every byte is a command word, split into the two
// nibbles. More significant nibble is type of command, lower one
// selects the command
//
// Note: Except for Master Reset and Enter UART Mode, none of these are
//       supported at the moment. They will have to be implemented as
//       IOCTL calls to the /dev/midi00 driver
    switch (value >> 4)
	{
	case 0:	switch (value & 15) {	// MIDI commands
		case 1: writelog("MIDI: Stop");
                        success=1;
			break;
		case 2: writelog("MIDI: Start");
                        success=1;
			break;
		case 3: writelog("MIDI: Continue");
                        success=1;
			break;
		}
		break;
	case 1:	switch (value & 15) {	// system commands
		case 5:	writelog("SYST: Stop device");
                        success=1;
			break;
		}
		break;
	case 3: switch (value & 15) {	// mode commands
		case 4: writelog("MODE: Return timing bytes in stop mode");
                        success=1;
			break;
		case 5: writelog("MODE: Enable mode messages");
                        success=1;
			break;
		case 8: writelog("MODE: Enable system common messages");
                        success=1;
			break;
		case 9: writelog("MODE: Enable realtime messages");
                        success=1;
			break;
		case 15:writelog("MODE: UART mode on");
			BX_MPU401_THIS s.uartmode=1;
			success=1;
		}
		break;
	case 8:	switch (value & 15) {	//  controller commands
		case 0:	writelog("CTRL: Use internal MIDI timing");
                        success=1;
			break;
		case 1:	writelog("CTRL: Use FSK timing");
                        success=1;
			break;
		case 2:	writelog("CTRL: Use external MIDI timing");
                        success=1;
			break;
		case 3:	writelog("CTRL: Enable metronome");
                        success=1;
			break;
		case 4:	writelog("CTRL: Disable metronome");
                        success=1;
			break;
		case 7:	writelog("CTRL: Enable pitch and controller");
                        success=1;
			break;
		case 10:writelog("CTRL: Disable data in stop mode");
                        success=1;
			break;
		case 11:writelog("CTRL: Enable data in stop mode");
                        success=1;
			break;
		}
		break;
	case 9:	switch (value & 15) {	// Clock commands
		case 4:	writelog("CLCK: Disable clock to host");
                        success=1;
			break;
		case 5: writelog("CLCK: Enable clock to host");
                        success=1;
                        break;
		}
		break;
	case 10:switch (value & 15) {	// Version commands
                case 12:writelog("VERS: Return version number");
                        success=1;
                        break;
		case 13:writelog("VERS: Return revision number");
			success=1;
			break;
		}
		break;
        case 12:  // Timebase commands
                writelog("TBSE: Set timebase");
                success=1;
                break;
        case 14:switch (value & 15) {   // Tempo commands
                case 0: writelog("TEMP: Set tempo");
                        success=1;
                        break;
                case 4: writelog("TEMP: Set metronome clocks per click");
                        success=1;
                        break;
                case 6: writelog("TEMP: Set metronome beats per minutes");
                        success=1;
                        break;
                case 7: writelog("TEMP: Send clocks to host");
                        success=1;
                        break;
                }
                break;
        case 15:switch (value & 15) {   // Special commands
                case 15:writelog("SPCL: Master reset of device");
			BX_MPU401_THIS s.uartmode=0;
                        success=1;
                        break;
                }
                break;
	}
    if (success == 0)
	writelog("Unknown command");

// Need to put an MPU_ACK into the data port
    if (OUTBUF put(MPU_ACK) == 0)
	writelog("MPU_ACK error - output buffer full");
    }
  else if (address == DATAPORT)
    {
// data port: This is where the midi streams comes from.
#if BX_MPU401_LOG > 1
	writelog("write to data port, value %02x", value);
#endif
	writedata( (Bit8u) value);
    }
  else
	writelog("write to unsupported port!");

  /* MPU401 */
}

void bx_mpu401_c::writelog(const char *str, ...)
{
// append a line to the log file, if desired
#if BX_MPU401_LOG > 0
	fprintf(LOGFILE, "MPU401: ");
	va_list ap;
	va_start(ap, str);
	vfprintf(LOGFILE, str, ap);
	va_end(ap);
	fprintf(LOGFILE, "\n");
	fflush(LOGFILE);
#endif
}

void bx_mpu401_c::writedata(const Bit8u value)
{
fputc(value, DATAFILE);
fflush(DATAFILE);	// to start playing immediately
}

// the round-robin FIFO buffers of the MPU401
bx_mpu401_buffer::bx_mpu401_buffer()
{
  length = 0;		// total bytes in buffer
  head = 0;		// pointer to next slot available for new data
  tail = 0;		// pointer to next slot to be read from
  buffer = NULL;	// pointer to the actual data
}

void bx_mpu401_buffer::init(int bufferlen)
{
  if (buffer != NULL)	// Was it initialized before?
    delete buffer;

  length = bufferlen;
  buffer = new Bit8u[length];
  if (buffer == NULL)
    length = 0;		// This will be checked later
  head = 0;		// Reset the pointers
  tail = 0;
}

bx_mpu401_buffer::~bx_mpu401_buffer(void)
{
  if (buffer != NULL)
    delete buffer;
}

// This puts one byte into the buffer
Boolean bx_mpu401_buffer::put(Bit8u data)
{
  if (full() != 0)
    return 0;	// buffer full

  buffer[head++] = data;	// Write data, and increase write pointer
  head %= length;		// wrap it around so it stays inside the data

  return 1;	// put was successful
}

// This returns if the buffer is full, i.e. if a put will fail
Boolean bx_mpu401_buffer::full(void)
{
  if (length == 0)
    return 1;   // not initialized

  if ( ((head + 1) % length) == tail)
    return 1;   // buffer full

  return 0;	// buffer has some space left
}

// This reads the next available byte from the buffer
Boolean bx_mpu401_buffer::get(Bit8u *data)
{
  if (empty() != 0)
    return 0;	// buffer empty

  (*data) = buffer[tail++];	// read data and increase read pointer
  tail %= length;		// and wrap it around to stay inside the data

  return 1;	// get was successful
}

// This returns if the buffer is empty, i.e. if a get will fail
Boolean bx_mpu401_buffer::empty(void)
{
  if (length == 0)
    return 1;   // not inialized

  if (head == tail)
    return 1;   // buffer empty

  return 0;	// buffer contains data
}
